package explorviz.hpc_monitoring.filter.reconstruction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.disruptor.RecordArrayEvent;
import explorviz.hpc_monitoring.disruptor.RecordEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.filter.reduction.TracePatternSummarizationFilter;
import explorviz.hpc_monitoring.reader.TimeProvider;
import explorviz.hpc_monitoring.record.HostApplicationMetaData;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.TerminateRecord;
import explorviz.hpc_monitoring.record.TimedPeriodRecord;
import explorviz.hpc_monitoring.record.event.AbstractOperationEvent;
import explorviz.hpc_monitoring.record.trace.Trace;

public final class TraceReconstructionFilter implements EventHandler<RecordArrayEvent> {
	private static final int OUTPUT_MESSAGE_BUFFER_SIZE = 256;

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Reconstructed traces/sec");

	private final long maxTraceTimeout;

	private final Map<Long, TraceBuffer> traceId2trace = new ConcurrentSkipListMap<Long, TraceBuffer>();

	private final RingBuffer<RecordArrayEvent> ringBuffer;
	private IRecord[] outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	public TraceReconstructionFilter(final long maxTraceTimeout,
			final EventHandler<RecordEvent> endReceiver) {
		this.maxTraceTimeout = maxTraceTimeout;

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordArrayEvent> disruptor = new Disruptor<RecordArrayEvent>(
				RecordArrayEvent.EVENT_FACTORY, 32, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<RecordArrayEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new TracePatternSummarizationFilter(1 * 1000, endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();
	}

	private void periodicTimeSignal(final long timestamp) {
		checkForTimeouts(timestamp);
		flushOutputBuffer();
	}

	private void checkForTimeouts(final long timestamp) {
		final long traceTimeout = timestamp - maxTraceTimeout;
		final List<Long> traceIdsToRemove = new ArrayList<Long>();

		for (final Entry<Long, TraceBuffer> entry : traceId2trace.entrySet()) {
			final TraceBuffer traceBuffer = entry.getValue();
			if (traceBuffer.getMaxLoggingTimestamp() <= traceTimeout) {
				sendOutInvalidTrace(traceBuffer.toTrace());
				traceIdsToRemove.add(entry.getKey());
			}
		}

		for (final Long traceIdToRemove : traceIdsToRemove) {
			traceId2trace.remove(traceIdToRemove);
		}
	}

	private void sendOutValidTrace(final Trace trace) {
		putInRingBuffer(trace);
	}

	private void sendOutInvalidTrace(final Trace trace) {
		putInRingBuffer(trace);
		System.out.println("Invalid trace: " + trace.getTraceEvents()[0].getTraceId());
	}

	private void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);
		outputBuffer[outputBufferIndex++] = message;

		if (outputBufferIndex == OUTPUT_MESSAGE_BUFFER_SIZE) {
			flushOutputBuffer();
		}
	}

	private void flushOutputBuffer() {
		if (outputBufferIndex > 0) {
			final long hiseq = ringBuffer.next();
			final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
			valueEvent.setValues(outputBuffer);
			valueEvent.setValuesLength(outputBufferIndex);
			ringBuffer.publish(hiseq);

			outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
			outputBufferIndex = 0;
		}
	}

	@Override
	public void onEvent(final RecordArrayEvent event, final long sequence, final boolean endOfBatch)
			throws Exception {
		final IRecord[] values = event.getValues();
		final int valuesLength = event.getValuesLength();

		for (int i = 0; i < valuesLength; i++) {
			final IRecord record = values[i];
			if (record instanceof AbstractOperationEvent) {
				final AbstractOperationEvent abstractOperationEvent = ((AbstractOperationEvent) record);

				final long traceId = abstractOperationEvent.getTraceId();
				final TraceBuffer traceBuffer = getBufferForTraceId(
						abstractOperationEvent.getTraceId(), event.getMetadata());
				traceBuffer.insertEvent(abstractOperationEvent);

				if (traceBuffer.isFinished()) {
					traceId2trace.remove(traceId);
					sendOutValidTrace(traceBuffer.toTrace());
				}
			} else if (record instanceof TimedPeriodRecord) {
				periodicTimeSignal(TimeProvider.getCurrentTimestamp());
			} else if (record instanceof TerminateRecord) {
				terminate();
			}
		}
	}

	private TraceBuffer getBufferForTraceId(final long traceId,
			final HostApplicationMetaData metadata) {
		TraceBuffer traceBuffer = traceId2trace.get(traceId);
		if (traceBuffer == null) {
			traceBuffer = new TraceBuffer();
			traceBuffer.setTrace(metadata);
			traceId2trace.put(traceId, traceBuffer);
		}
		return traceBuffer;
	}

	private void terminate() {
		for (final TraceBuffer entry : traceId2trace.values()) {
			sendOutInvalidTrace(entry.toTrace());
		}
		traceId2trace.clear();
	}
}